@@ -37,6 +37,7 @@ use std::{fs, os::unix::prelude::FileExt, path::PathBuf};
 
 use std::io::prelude::*;
 
+use fs_extra::dir;
 use goblin::elf::Elf;
 use serde::{Deserialize, Serialize};
 use sys_mount::{Mount, MountFlags, Unmount, UnmountFlags};
@@ -61,6 +62,13 @@ pub struct FwConfig {
     files: Vec<FwFile>,
 }
 
+#[derive(Deserialize)]
+pub struct FwFolder {
+    partition: String,
+    destination: String,
+    folders: Vec<FwFile>,
+}
+
 #[derive(Deserialize)]
 pub struct DumpConfig {
     partition: String,
@@ -72,12 +80,14 @@ pub struct DumpConfig {
 pub struct Config {
     dynpart: Option<String>,
     firmware: Vec<FwConfig>,
+    folders: Option<Vec<FwFolder>>,
     partdump: Option<Vec<DumpConfig>>,
 }
 
 #[derive(Serialize, Deserialize)]
 pub struct Status {
     pub files: Vec<String>,
+    pub folders: Option<Vec<String>>,
 }
 
 #[derive(Serialize, Deserialize)]
@@ -89,30 +99,34 @@ pub struct OldStatus {
 fn mount_part(part: &str, mountpath: &PathBuf) -> Result<Mount, Error> {
     let _res = fs::DirBuilder::new().recursive(true).create(mountpath);
     let flags = MountFlags::RDONLY;
+    let part_a = format!("{}_a", part);
+    let part_b = format!("{}_b", part);
 
     let mut srcpath = PathBuf::from("/dev/mapper").join(part);
     if !srcpath.exists() {
-        srcpath.set_file_name(format!("{}_a", part));
+        srcpath.set_file_name(&part_a);
     }
     if !srcpath.exists() {
         srcpath = PathBuf::from(PARTLABEL_DIR).join(part);
     }
     if !srcpath.exists() {
-        srcpath.set_file_name(format!("{}_a", part));
+        srcpath.set_file_name(&part_a);
     }
 
+    debug!("Attempting to mount {}", srcpath.display());
+
     match Mount::builder().flags(flags).mount(&srcpath, mountpath) {
         Ok(m) => Ok(m),
         Err(e) => {
-            if srcpath.ends_with("_a") {
-                eprintln!(
+            if srcpath.ends_with(&part_a) {
+                error!(
                     "Unable to mount {} on {}: {}",
                     srcpath.display(),
                     mountpath.display(),
                     e
                 );
-                srcpath.set_file_name(format!("{}_b", part));
-                println!("Mounting {} on {}", srcpath.display(), mountpath.display());
+                srcpath.set_file_name(&part_b);
+                info!("Mounting {} on {}", srcpath.display(), mountpath.display());
                 Mount::builder().flags(flags).mount(&srcpath, mountpath)
             } else {
                 Err(e)
@@ -125,7 +139,7 @@ fn squash_file(inpath: &PathBuf, outpath: &PathBuf) -> Result<(), Error> {
     let buffer = match fs::read(inpath) {
         Ok(buf) => buf,
         Err(e) => {
-            eprintln!("Unable to read {}: {}", inpath.display(), e);
+            error!("Unable to read {}: {}", inpath.display(), e);
             return Err(e);
         }
     };
@@ -133,7 +147,7 @@ fn squash_file(inpath: &PathBuf, outpath: &PathBuf) -> Result<(), Error> {
     let elf = match Elf::parse(buffer.as_slice()) {
         Ok(value) => value,
         Err(e) => {
-            eprintln!("Unable to parse {}: {}", inpath.display(), e);
+            error!("Unable to parse {}: {}", inpath.display(), e);
             return Err(Error::new(ErrorKind::InvalidData, e));
         }
     };
@@ -195,20 +209,23 @@ fn map_dynpart(part: &str) -> Result<(), Error> {
             ]),
         )
     } else {
-        let err_str = format!("Warning: unable to find super partition '{}'", part);
+        let err_str = format!("Unable to find super partition '{}'", part);
         Err(Error::new(ErrorKind::Other, err_str))
     }
 }
 
 pub fn process(config: Config) -> Result<Status, Error> {
     let mut files: Vec<String> = Vec::new();
+    let mut folders: Option<Vec<String>> = None;
 
     // Map the "super" partition if we expect one
     if let Some(part) = config.dynpart {
+        info!("Mounting {} as the 'super' partition", part);
         let mut success = false;
 
         for suffix in ["", "_a", "_b"] {
             let testpart = format!("{}{}", part, suffix);
+            debug!("Attempting to map dynpart {}", testpart);
             if map_dynpart(&testpart).is_ok() {
                 success = true;
                 break;
@@ -227,11 +244,7 @@ pub fn process(config: Config) -> Result<Status, Error> {
         let destpath = PathBuf::from("/lib/firmware/updates").join(entry.destination);
 
         if let Err(e) = fs::create_dir_all(&destpath) {
-            eprintln!(
-                "Warning: unable to create folder {}: {}",
-                destpath.display(),
-                e
-            );
+            warn!("Unable to create folder {}: {}", destpath.display(), e);
             continue;
         }
 
@@ -239,11 +252,15 @@ pub fn process(config: Config) -> Result<Status, Error> {
 
         match mount_part(entry.partition.as_str(), &mntpath) {
             Ok(m) => {
+                debug!(
+                    "Processing firmware files from partition {}",
+                    entry.partition.as_str()
+                );
                 for file in entry.files {
                     let origin = PathBuf::from(&mntpath).join(&entry.origin).join(&file.name);
                     if !origin.exists() {
-                        eprintln!(
-                            "Warning: unable to find {} on partition {}",
+                        warn!(
+                            "Unable to find {} on partition {}",
                             file.name, entry.partition
                         );
                         continue;
@@ -254,11 +271,14 @@ pub fn process(config: Config) -> Result<Status, Error> {
                         destination.set_file_name(&new_name);
                     }
 
+                    debug!("Copying firmware file {}", origin.display());
+
                     if file.name.ends_with(".mdt") {
+                        trace!("Squashing MDT file into MBN");
                         destination.set_extension("mbn");
                         if let Err(e) = squash_file(&origin, &destination) {
-                            eprintln!(
-                                "Warning: unable to squash {} to {}: {}",
+                            warn!(
+                                "Unable to squash {} to {}: {}",
                                 origin.display(),
                                 destination.display(),
                                 e
@@ -266,8 +286,8 @@ pub fn process(config: Config) -> Result<Status, Error> {
                             continue;
                         }
                     } else if let Err(e) = fs::copy(&origin, &destination) {
-                        eprintln!(
-                            "Warning: unable to copy {} to {}: {}",
+                        warn!(
+                            "Unable to copy {} to {}: {}",
                             origin.display(),
                             destination.display(),
                             e
@@ -285,21 +305,85 @@ pub fn process(config: Config) -> Result<Status, Error> {
         let _r = fs::remove_dir(mntpath);
     }
 
+    if let Some(dirs) = config.folders {
+        let options = dir::CopyOptions::new();
+        let mut folder_list = Vec::new();
+
+        for entry in dirs {
+            let destpath = PathBuf::from(entry.destination);
+
+            if let Err(e) = fs::create_dir_all(&destpath) {
+                warn!("Unable to create folder {}: {}", destpath.display(), e);
+                continue;
+            }
+
+            let mntpath = PathBuf::from("/tmp").join(&entry.partition);
+
+            match mount_part(entry.partition.as_str(), &mntpath) {
+                Ok(m) => {
+                    debug!(
+                        "Processing folders from partition {}",
+                        entry.partition.as_str()
+                    );
+                    for folder in entry.folders {
+                        let origin = PathBuf::from(&mntpath).join(&folder.name);
+                        if !origin.exists() {
+                            warn!(
+                                "Unable to find {} on partition {}",
+                                folder.name, entry.partition
+                            );
+                            continue;
+                        }
+
+                        debug!("Copying folder {}", origin.display());
+
+                        if let Err(e) = dir::copy(&origin, &destpath, &options) {
+                            warn!(
+                                "Unable to copy {} to {}: {}",
+                                origin.display(),
+                                destpath.display(),
+                                e
+                            );
+                            continue;
+                        }
+
+                        let mut destination =
+                            PathBuf::from(&destpath).join(origin.file_name().unwrap());
+                        if let Some(new_name) = folder.rename {
+                            let initial_folder = PathBuf::from(&destination);
+                            destination.set_file_name(&new_name);
+                            let _ = fs::rename(initial_folder, &destination);
+                        }
+                        folder_list.push(format!("{}", destination.display()));
+                    }
+                    let _res = m.unmount(UnmountFlags::empty());
+                }
+                Err(e) => return Err(e),
+            }
+
+            let _r = fs::remove_dir(mntpath);
+        }
+
+        if !folder_list.is_empty() {
+            folders = Some(folder_list);
+        }
+    }
+
     if let Some(dumps) = config.partdump {
         for entry in dumps {
+            debug!(
+                "Processing partition {} for raw dump",
+                entry.partition.as_str()
+            );
             let destpath = PathBuf::from("/lib/firmware/updates").join(&entry.destination);
             if let Err(e) = fs::create_dir_all(&destpath) {
-                eprintln!(
-                    "Warning: unable to create folder {}: {}",
-                    destpath.display(),
-                    e
-                );
+                warn!("Unable to create folder {}: {}", destpath.display(), e);
                 continue;
             }
 
             let origin = PathBuf::from(PARTLABEL_DIR).join(&entry.partition);
             if !origin.exists() {
-                eprintln!("Warning: unable to find partition {}", entry.partition);
+                warn!("Unable to find partition {}", entry.partition);
                 continue;
             }
 
@@ -315,5 +399,5 @@ pub fn process(config: Config) -> Result<Status, Error> {
         }
     }
 
-    Ok(Status { files })
+    Ok(Status { files, folders })
 }
